@@ -1,27 +1,27 @@
 // Copyright (C) 2016 Jolla Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
-#include <qvideoframe.h>
-#include <qvideosink.h>
-#include <QtCore/QDebug>
-#include <QtCore/QMap>
-#include <QtCore/QThread>
-#include <QtCore/QEvent>
-#include <QtCore/QCoreApplication>
-#include <QtCore/private/quniquehandle_p.h>
-
-#include <private/qfactoryloader_p.h>
-#include <qgstvideobuffer_p.h>
-#include <qgstreamervideosink_p.h>
-#include <qgst_debug_p.h>
-#include <qgstutils_p.h>
-
 #include "qgstvideorenderersink_p.h"
 
+#include <QtMultimedia/qvideoframe.h>
+#include <QtMultimedia/qvideosink.h>
+#include <QtCore/private/qfactoryloader_p.h>
+#include <QtCore/private/quniquehandle_p.h>
+#include <QtCore/qcoreapplication.h>
+#include <QtCore/qdebug.h>
+#include <QtCore/qdebug.h>
+#include <QtCore/qloggingcategory.h>
+#include <QtCore/qmap.h>
+#include <QtCore/qthread.h>
+#include <QtGui/qevent.h>
+
+#include <common/qgstvideobuffer_p.h>
+#include <common/qgstreamervideosink_p.h>
+#include <common/qgst_debug_p.h>
+#include <common/qgstutils_p.h>
+
 #include <gst/video/video.h>
 #include <gst/video/gstvideometa.h>
-#include <qloggingcategory.h>
-#include <qdebug.h>
 
 
 #include <rhi/qrhi.h>
@@ -34,8 +34,6 @@
 #include <gst/allocators/gstdmabuf.h>
 #endif
 
-//#define DEBUG_VIDEO_SURFACE_SINK
-
 static Q_LOGGING_CATEGORY(qLcGstVideoRenderer, "qt.multimedia.gstvideorenderer")
 
 QT_BEGIN_NAMESPACE
@@ -43,13 +41,18 @@ QT_BEGIN_NAMESPACE
 QGstVideoRenderer::QGstVideoRenderer(QGstreamerVideoSink *sink)
     : m_sink(sink), m_surfaceCaps(createSurfaceCaps(sink))
 {
+    QObject::connect(
+            sink, &QGstreamerVideoSink::aboutToBeDestroyed, this,
+            [this] {
+                QMutexLocker locker(&m_sinkMutex);
+                m_sink = nullptr;
+            },
+            Qt::DirectConnection);
 }
 
-QGstVideoRenderer::~QGstVideoRenderer()
-{
-}
+QGstVideoRenderer::~QGstVideoRenderer() = default;
 
-QGstCaps QGstVideoRenderer::createSurfaceCaps(QGstreamerVideoSink *sink)
+QGstCaps QGstVideoRenderer::createSurfaceCaps([[maybe_unused]] QGstreamerVideoSink *sink)
 {
     QGstCaps caps = QGstCaps::create();
 
@@ -115,91 +118,105 @@ const QGstCaps &QGstVideoRenderer::caps()
 bool QGstVideoRenderer::start(const QGstCaps& caps)
 {
     qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::start" << caps;
-    QMutexLocker locker(&m_mutex);
 
-    m_frameMirrored = false;
-    m_frameRotationAngle = QtVideo::Rotation::None;
-
-    if (m_active) {
-        m_flush = true;
-        m_stop = true;
+    {
+        m_frameRotationAngle = QtVideo::Rotation::None;
+        auto optionalFormatAndVideoInfo = caps.formatAndVideoInfo();
+        if (optionalFormatAndVideoInfo) {
+            std::tie(m_format, m_videoInfo) = std::move(*optionalFormatAndVideoInfo);
+        } else {
+            m_format = {};
+            m_videoInfo = {};
+        }
+        m_memoryFormat = caps.memoryFormat();
     }
 
-    m_startCaps = caps;
-
-    /*
-    Waiting for start() to be invoked in the main thread may block
-    if gstreamer blocks the main thread until this call is finished.
-    This situation is rare and usually caused by setState(Null)
-    while pipeline is being prerolled.
-
-    The proper solution to this involves controlling gstreamer pipeline from
-    other thread than video surface.
-
-    Currently start() fails if wait() timed out.
-    */
-    if (!waitForAsyncEvent(&locker, &m_setupCondition, 1000) && !m_startCaps.isNull()) {
-        qWarning() << "Failed to start video surface due to main thread blocked.";
-        m_startCaps = {};
-    }
-
-    return m_active;
+    return true;
 }
 
 void QGstVideoRenderer::stop()
 {
-    QMutexLocker locker(&m_mutex);
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::stop";
 
-    if (!m_active)
+    QMetaObject::invokeMethod(this, [this] {
+        m_currentState.buffer = {};
+        m_sink->setVideoFrame(QVideoFrame{});
         return;
-
-    m_flush = true;
-    m_stop = true;
-
-    m_startCaps = {};
-
-    waitForAsyncEvent(&locker, &m_setupCondition, 500);
+    });
 }
 
 void QGstVideoRenderer::unlock()
 {
-    QMutexLocker locker(&m_mutex);
-
-    m_setupCondition.wakeAll();
-    m_renderCondition.wakeAll();
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::unlock";
 }
 
-bool QGstVideoRenderer::proposeAllocation(GstQuery *query)
+bool QGstVideoRenderer::proposeAllocation(GstQuery *)
 {
-    Q_UNUSED(query);
-    QMutexLocker locker(&m_mutex);
-    return m_active;
-}
-
-void QGstVideoRenderer::flush()
-{
-    QMutexLocker locker(&m_mutex);
-
-    m_flush = true;
-    m_renderBuffer = nullptr;
-    m_renderCondition.wakeAll();
-
-    notify();
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::proposeAllocation";
+    return true;
 }
 
 GstFlowReturn QGstVideoRenderer::render(GstBuffer *buffer)
 {
-    QMutexLocker locker(&m_mutex);
     qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::render";
 
-    m_renderReturn = GST_FLOW_OK;
-    m_renderBuffer = buffer;
+    GstVideoCropMeta *meta = gst_buffer_get_video_crop_meta(buffer);
+    if (meta) {
+        QRect vp(meta->x, meta->y, meta->width, meta->height);
+        if (m_format.viewport() != vp) {
+            qCDebug(qLcGstVideoRenderer)
+                    << Q_FUNC_INFO << " Update viewport on Metadata: [" << meta->height << "x"
+                    << meta->width << " | " << meta->x << "x" << meta->y << "]";
+            // Update viewport if data is not the same
+            m_format.setViewport(vp);
+        }
+    }
 
-    waitForAsyncEvent(&locker, &m_renderCondition, 300);
+    RenderBufferState state{
+        .buffer = QGstBufferHandle{ buffer, QGstBufferHandle::NeedsRef },
+        .format = m_format,
+        .memoryFormat = m_memoryFormat,
+        .mirrored = m_frameMirrored,
+        .rotationAngle = m_frameRotationAngle,
+    };
 
-    m_renderBuffer = nullptr;
+    qCDebug(qLcGstVideoRenderer) << "    sending video frame";
 
-    return m_renderReturn;
+    QMetaObject::invokeMethod(this, [this, state = std::move(state)]() mutable {
+        if (state == m_currentState) {
+            // same buffer received twice
+            if (!m_sink || !m_sink->inStoppedState())
+                return;
+
+            qCDebug(qLcGstVideoRenderer) << "    showing empty video frame";
+            m_currentVideoFrame = {};
+            m_sink->setVideoFrame(m_currentVideoFrame);
+            m_currentState = {};
+            return;
+        }
+
+        QGstVideoBuffer *videoBuffer = new QGstVideoBuffer{
+            state.buffer, m_videoInfo, m_sink, state.format, state.memoryFormat,
+        };
+        QVideoFrame frame(videoBuffer, state.format);
+        QGstUtils::setFrameTimeStampsFromBuffer(&frame, state.buffer.get());
+        frame.setMirrored(state.mirrored);
+        frame.setRotation(state.rotationAngle);
+        m_currentVideoFrame = std::move(frame);
+        m_currentState = std::move(state);
+
+        if (!m_sink)
+            return;
+
+        if (m_sink->inStoppedState()) {
+            qCDebug(qLcGstVideoRenderer) << "    showing empty video frame";
+            m_currentVideoFrame = {};
+        }
+
+        m_sink->setVideoFrame(m_currentVideoFrame);
+    });
+
+    return GST_FLOW_OK;
 }
 
 bool QGstVideoRenderer::query(GstQuery *query)
@@ -212,6 +229,10 @@ bool QGstVideoRenderer::query(GstQuery *query)
         if (strcmp(type, "gst.gl.local_context") != 0)
             return false;
 
+        QMutexLocker locker(&m_sinkMutex);
+        if (!m_sink)
+            return false;
+
         auto *gstGlContext = m_sink->gstGlLocalContext();
         if (!gstGlContext)
             return false;
@@ -228,9 +249,22 @@ bool QGstVideoRenderer::query(GstQuery *query)
 
 void QGstVideoRenderer::gstEvent(GstEvent *event)
 {
-    if (GST_EVENT_TYPE(event) != GST_EVENT_TAG)
-        return;
+    switch (GST_EVENT_TYPE(event)) {
+    case GST_EVENT_TAG:
+        qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::gstEvent: Tag";
+        return gstEventHandleTag(event);
+    case GST_EVENT_EOS:
+        qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::gstEvent: EOS";
+        return gstEventHandleEOS(event);
 
+    default:
+        qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::gstEvent: unhandled event - " << event;
+        return;
+    }
+}
+
+void QGstVideoRenderer::gstEventHandleTag(GstEvent *event)
+{
     GstTagList *taglist = nullptr;
     gst_event_parse_tag(event, &taglist);
     if (!taglist)
@@ -257,152 +291,28 @@ void QGstVideoRenderer::gstEvent(GstEvent *event)
         rotationAngle = (180 + atoi(value.get() + flipRotateLen)) % 360;
     }
 
-    QMutexLocker locker(&m_mutex);
     m_frameMirrored = mirrored;
     switch (rotationAngle) {
-    case 0: m_frameRotationAngle = QtVideo::Rotation::None; break;
-    case 90: m_frameRotationAngle = QtVideo::Rotation::Clockwise90; break;
-    case 180: m_frameRotationAngle = QtVideo::Rotation::Clockwise180; break;
-    case 270: m_frameRotationAngle = QtVideo::Rotation::Clockwise270; break;
-    default: m_frameRotationAngle = QtVideo::Rotation::None;
+    case 0:
+        m_frameRotationAngle = QtVideo::Rotation::None;
+        break;
+    case 90:
+        m_frameRotationAngle = QtVideo::Rotation::Clockwise90;
+        break;
+    case 180:
+        m_frameRotationAngle = QtVideo::Rotation::Clockwise180;
+        break;
+    case 270:
+        m_frameRotationAngle = QtVideo::Rotation::Clockwise270;
+        break;
+    default:
+        m_frameRotationAngle = QtVideo::Rotation::None;
     }
 }
 
-bool QGstVideoRenderer::event(QEvent *event)
+void QGstVideoRenderer::gstEventHandleEOS(GstEvent *)
 {
-    if (event->type() == QEvent::UpdateRequest) {
-        QMutexLocker locker(&m_mutex);
-
-        if (m_notified) {
-            while (handleEvent(&locker)) {}
-            m_notified = false;
-        }
-        return true;
-    }
-
-    return QObject::event(event);
-}
-
-bool QGstVideoRenderer::handleEvent(QMutexLocker<QMutex> *locker)
-{
-    if (m_flush) {
-        m_flush = false;
-        if (m_active) {
-            locker->unlock();
-
-            if (m_sink && !m_flushed)
-                m_sink->setVideoFrame(QVideoFrame());
-            m_flushed = true;
-            locker->relock();
-        }
-    } else if (m_stop) {
-        m_stop = false;
-
-        if (m_active) {
-            m_active = false;
-            m_flushed = true;
-        }
-    } else if (!m_startCaps.isNull()) {
-        Q_ASSERT(!m_active);
-
-        auto startCaps = m_startCaps;
-        m_startCaps = {};
-
-        if (m_sink) {
-            locker->unlock();
-
-            m_flushed = true;
-            auto optionalFormatAndVideoInfo = startCaps.formatAndVideoInfo();
-            if (optionalFormatAndVideoInfo) {
-                std::tie(m_format, m_videoInfo) = std::move(*optionalFormatAndVideoInfo);
-            } else {
-                m_format = {};
-                m_videoInfo = {};
-            }
-
-            memoryFormat = startCaps.memoryFormat();
-
-            locker->relock();
-            m_active = m_format.isValid();
-        } else if (m_active) {
-            m_active = false;
-            m_flushed = true;
-        }
-
-    } else if (m_renderBuffer) {
-        GstBuffer *buffer = m_renderBuffer;
-        m_renderBuffer = nullptr;
-        m_renderReturn = GST_FLOW_ERROR;
-
-        qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::handleEvent(renderBuffer)" << m_active << m_sink;
-        if (m_active && m_sink) {
-            gst_buffer_ref(buffer);
-
-            locker->unlock();
-
-            m_flushed = false;
-
-            GstVideoCropMeta *meta = gst_buffer_get_video_crop_meta(buffer);
-            if (meta) {
-                QRect vp(meta->x, meta->y, meta->width, meta->height);
-                if (m_format.viewport() != vp) {
-                    qCDebug(qLcGstVideoRenderer) << Q_FUNC_INFO << " Update viewport on Metadata: [" << meta->height << "x" << meta->width << " | " << meta->x << "x" << meta->y << "]";
-                    // Update viewport if data is not the same
-                    m_format.setViewport(vp);
-                }
-            }
-
-            if (m_sink->inStoppedState()) {
-                qCDebug(qLcGstVideoRenderer) << "    sending empty video frame";
-                m_sink->setVideoFrame(QVideoFrame());
-            } else {
-                QGstVideoBuffer *videoBuffer = new QGstVideoBuffer(buffer, m_videoInfo, m_sink, m_format, memoryFormat);
-                QVideoFrame frame(videoBuffer, m_format);
-                QGstUtils::setFrameTimeStamps(&frame, buffer);
-                frame.setMirrored(m_frameMirrored);
-                frame.setRotationAngle(QVideoFrame::RotationAngle(m_frameRotationAngle));
-
-                qCDebug(qLcGstVideoRenderer) << "    sending video frame";
-                m_sink->setVideoFrame(frame);
-            }
-
-            gst_buffer_unref(buffer);
-
-            locker->relock();
-
-            m_renderReturn = GST_FLOW_OK;
-        }
-
-        m_renderCondition.wakeAll();
-    } else {
-        m_setupCondition.wakeAll();
-
-        return false;
-    }
-    return true;
-}
-
-void QGstVideoRenderer::notify()
-{
-    if (!m_notified) {
-        m_notified = true;
-        QCoreApplication::postEvent(this, new QEvent(QEvent::UpdateRequest));
-    }
-}
-
-bool QGstVideoRenderer::waitForAsyncEvent(
-        QMutexLocker<QMutex> *locker, QWaitCondition *condition, unsigned long time)
-{
-    if (QThread::currentThread() == thread()) {
-        while (handleEvent(locker)) {}
-        m_notified = false;
-
-        return true;
-    }
-
-    notify();
-
-    return condition->wait(&m_mutex, time);
+    stop();
 }
 
 static GstVideoSinkClass *gvrs_sink_parent_class;
@@ -416,8 +326,6 @@ QGstVideoRendererSink *QGstVideoRendererSink::createSink(QGstreamerVideoSink *si
     QGstVideoRendererSink *gstSink = reinterpret_cast<QGstVideoRendererSink *>(
             g_object_new(QGstVideoRendererSink::get_type(), nullptr));
 
-    g_signal_connect(G_OBJECT(gstSink), "notify::show-preroll-frame", G_CALLBACK(handleShowPrerollChange), gstSink);
-
     return gstSink;
 }
 
@@ -442,16 +350,8 @@ GType QGstVideoRendererSink::get_type()
         nullptr                                            // value_table
     };
 
-    static const GType type = []() {
-        const auto result = g_type_register_static(
-                GST_TYPE_VIDEO_SINK, "QGstVideoRendererSink", &info, GTypeFlags(0));
-
-        // Register the sink type to be used in custom piplines.
-        // When surface is ready the sink can be used.
-        gst_element_register(nullptr, "qtvideosink", GST_RANK_PRIMARY, result);
-
-        return result;
-    }();
+    static const GType type = g_type_register_static(GST_TYPE_VIDEO_SINK, "QGstVideoRendererSink",
+                                                     &info, GTypeFlags(0));
 
     return type;
 }
@@ -521,41 +421,9 @@ void QGstVideoRendererSink::finalize(GObject *object)
     G_OBJECT_CLASS(gvrs_sink_parent_class)->finalize(object);
 }
 
-void QGstVideoRendererSink::handleShowPrerollChange(GObject *o, GParamSpec *p, gpointer d)
-{
-    Q_UNUSED(o);
-    Q_UNUSED(p);
-    QGstVideoRendererSink *sink = reinterpret_cast<QGstVideoRendererSink *>(d);
-
-    gboolean showPrerollFrame = true; // "show-preroll-frame" property is true by default
-    g_object_get(G_OBJECT(sink), "show-preroll-frame", &showPrerollFrame, nullptr);
-
-    if (!showPrerollFrame) {
-        GstState state = GST_STATE_VOID_PENDING;
-        GstClockTime timeout = 10000000; // 10 ms
-        gst_element_get_state(GST_ELEMENT(sink), &state, nullptr, timeout);
-        // show-preroll-frame being set to 'false' while in GST_STATE_PAUSED means
-        // the QMediaPlayer was stopped from the paused state.
-        // We need to flush the current frame.
-        if (state == GST_STATE_PAUSED)
-            sink->renderer->flush();
-    }
-}
-
 GstStateChangeReturn QGstVideoRendererSink::change_state(
         GstElement *element, GstStateChange transition)
 {
-    QGstVideoRendererSink *sink = reinterpret_cast<QGstVideoRendererSink *>(element);
-
-    gboolean showPrerollFrame = true; // "show-preroll-frame" property is true by default
-    g_object_get(G_OBJECT(element), "show-preroll-frame", &showPrerollFrame, nullptr);
-
-    // If show-preroll-frame is 'false' when transitioning from GST_STATE_PLAYING to
-    // GST_STATE_PAUSED, it means the QMediaPlayer was stopped.
-    // We need to flush the current frame.
-    if (transition == GST_STATE_CHANGE_PLAYING_TO_PAUSED && !showPrerollFrame)
-        sink->renderer->flush();
-
     return GST_ELEMENT_CLASS(gvrs_sink_parent_class)->change_state(element, transition);
 }
 
@@ -573,20 +441,16 @@ GstCaps *QGstVideoRendererSink::get_caps(GstBaseSink *base, GstCaps *filter)
 gboolean QGstVideoRendererSink::set_caps(GstBaseSink *base, GstCaps *gcaps)
 {
     VO_SINK(base);
-
     auto caps = QGstCaps(gcaps, QGstCaps::NeedsRef);
 
     qCDebug(qLcGstVideoRenderer) << "set_caps:" << caps;
 
     if (caps.isNull()) {
         sink->renderer->stop();
-
         return TRUE;
-    } else if (sink->renderer->start(caps)) {
-        return TRUE;
-    } else {
-        return FALSE;
     }
+
+    return sink->renderer->start(caps);
 }
 
 gboolean QGstVideoRendererSink::propose_allocation(GstBaseSink *base, GstQuery *query)
@@ -632,5 +496,3 @@ gboolean QGstVideoRendererSink::event(GstBaseSink *base, GstEvent * event)
 }
 
 QT_END_NAMESPACE
-
-#include "moc_qgstvideorenderersink_p.cpp"
